/*
 * Copyright 2011-2012 (C) Christian Franke <nobody at nowhere dot ws>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with This program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include <avr/interrupt.h>
#include <avr/io.h>
#include <stdint.h>

/*
 * +-A-+
 * F   B
 * +-G-+
 * E   C
 * +-D-+ (DP)
 */

#define SEG_A (1<<0)
#define SEG_B (1<<1)
#define SEG_C (1<<2)
#define SEG_D (1<<3)
#define SEG_E (1<<4)
#define SEG_F (1<<5)
#define SEG_G (1<<6)
#define SEG_DP (1<<7)

/* 
 * This table is used to map decimal digits to 7 segment display settings.
 */
uint8_t number_map[] = {
	[0] = SEG_A | SEG_B | SEG_C | SEG_D | SEG_E | SEG_F,
	[1] = SEG_B | SEG_C,
	[2] = SEG_A | SEG_B | SEG_D | SEG_E | SEG_G,
	[3] = SEG_A | SEG_B | SEG_C | SEG_D | SEG_G,
	[4] = SEG_B | SEG_C | SEG_F | SEG_G,
	[5] = SEG_A | SEG_C | SEG_D | SEG_F | SEG_G,
	[6] = SEG_A | SEG_C | SEG_D | SEG_E | SEG_F | SEG_G,
	[7] = SEG_A | SEG_B | SEG_C,
	[8] = SEG_A | SEG_B | SEG_C | SEG_D | SEG_E | SEG_F | SEG_G,
	[9] = SEG_A | SEG_B | SEG_C | SEG_D | SEG_F | SEG_G
};

uint8_t minutes_left = 60;
uint8_t seconds_left = 0;

uint8_t multiplex_offset = 3;

ISR(TIMER0_OVF_vect)
{
	/* 
	 * This handler does the multiplexed display part. It switches
	 * the display to the next digit every time it it triggered 
	 */
	multiplex_offset++;
	if (multiplex_offset > 3)
		multiplex_offset = 0;

	/* turn all digits off */
	PORTD = 0x00;\

	/* load the next number */
	switch (multiplex_offset) {
	case 0:
		PORTB = number_map[seconds_left % 10];
		break;
	case 1:
		PORTB = number_map[seconds_left / 10];
		break;
	case 2:
		PORTB = number_map[minutes_left % 10] | SEG_DP;
		break;
	case 3:
		PORTB = number_map[(minutes_left / 10) % 10];
		break;
	}

	/* turn the next digit on */
	/*
	 * The left-most digit is digit 0 while the right-most digit
	 * is digit 3. While writing this code I thought it was the
	 * other way as documentation was somewhat lacky in that point.
	 *
	 * The quick fix to account for this is to substract the
	 * multiplex_offset from 3, this means for multiplex_offset == 0
	 * we actually show the real digit 3, which is the right-most
	 * digit and for multiplex_offset == 3 we show the real digit 0,
	 * which is the left-most digit.
	 */
	PORTD = (1 << (3 - multiplex_offset));
}

ISR(TIMER1_COMPA_vect)
{
	if (seconds_left) {
		seconds_left--;
		return;
	}

	if (minutes_left) {
		seconds_left = 59;
		minutes_left--;
		return;
	}
}

int main()
{
	PORTB = 0x00; /* Clear the segment control port */
	DDRB = 0xff; /* Set segment control port on output */

	PORTD = 0x00; /* Clear the digit controls */
	DDRD = 0x0f; /* Set the digit control pins on output */

	TIMSK = (1 << TOIE0); /* Enable Timer0 overflow interrupt */
	TCCR0B = (1 << CS01) | (1 << CS00); /* Multiplex with ~488 Hz (F/64/256)*/

	TIMSK |= (1 << OCIE1A); /* Enable Timer1 match A interrupt */
	TCCR1B = (1 << WGM12); /* set CTC mode for clock timer */

	OCR1A = 31249; /* set frequency for output compare to one Hertz */
	/* If your board is run with rc-oscillator, this value
	 * should be calibrated.
	 *
	 * Run the clock for some time you measure with a stop watch. (20
	 * minutes or longer should be okay, depending on the accuracy you
	 * need - it never will be very good with rc-oscillator :/ )
	 * Count/read the number of ticks the clock did in that time.
	 *
	 * OCR1A should be set to: (31250 * clock-ticks / real-seconds) - 1
	 *                            ^- this value is the (OCR1A + 1) of the
	 *                               clock while in the test
	 */

	TCCR1B |= (1 << CS12); /* set prescaler to /256 */

	sei(); /* Enable interrupts */

	for (;;); /* Everything happens in the interrupts, so we can idle here... */
}
